################################################################################
#
# Copyright 2023 ByteDance Ltd. and/or its affiliates. All rights reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
################################################################################

import torch
import sys
from contextlib import nullcontext, contextmanager


def _get_torch_fp8_dtypes():
    _torch_fp8_dtypes = []
    try:
        # from v2.1.0
        _torch_fp8_dtypes.append(torch.float8_e4m3fn)
        _torch_fp8_dtypes.append(torch.float8_e5m2)
        # from v2.2.0
        _torch_fp8_dtypes.append(torch.float8_e4m3fnuz)
        _torch_fp8_dtypes.append(torch.float8_e5m2fnuz)
    except Exception:
        pass
    return _torch_fp8_dtypes


def get_arch():
    properties = torch.cuda.get_device_properties(torch.cuda.current_device())
    major = properties.major
    minor = properties.minor
    return major * 10 + minor


def estimate_gemm_sol_time_ms(M: int, N: int, K: int, dtype=torch.bfloat16):
    """refer to this: https://arnon.dk/matching-sm-architectures-arch-and-gencode-for-various-nvidia-cards/"""

    def _get_flops():
        device_name = torch.cuda.get_device_name(
            torch.cuda.current_device()
        )  # arch is not a good idea. using device name is better.
        is_fp16 = dtype in [torch.bfloat16, torch.float16]
        is_fp8 = is_fp8_dtype(dtype)
        assert is_fp16 or is_fp8
        # https://www.nvidia.com/content/dam/en-zz/Solutions/Data-Center/a100/pdf/nvidia-a100-datasheet-nvidia-us-2188504-web.pdf
        # "NVIDIA A100 80GB PCIe" or "NVIDIA A100-SXM4-80GB" or "A100-SXM4-40GB"
        if device_name.find("A100") >= 0 or device_name.find("A800") >= 0:
            assert is_fp16
            return 312
        # https://resources.nvidia.com/en-us-gpu-resources/a10-datasheet-nvidia?lx=CPwSfP&ncid=no-ncid
        if device_name == "NVIDIA A10":  # No doc from NVIDIA
            return 125 if is_fp16 else 250
        if device_name == "NVIDIA A30":  # No doc from NVIDIA
            return 165 if is_fp16 else 330
        if device_name == "NVIDIA L20":  # No doc from NVIDIA
            return 119 if is_fp16 else 239
        # https://www.nvidia.com/en-us/data-center/l4/
        if device_name == "NVIDIA L4":
            return 121 if is_fp16 else 242
        # https://images.nvidia.com/content/Solutions/data-center/vgpu-L40-datasheet.pdf
        if device_name == "NVIDIA L40":
            return 181 if is_fp16 else 362
        # https://www.nvidia.com/en-us/data-center/l40s/
        if device_name == "NVIDIA L40S":
            return 366 if is_fp16 else 733
        # https://www.nvidia.com/en-us/data-center/h100/
        if device_name.find("H100") >= 0 or device_name.find("H800") >= 0:
            return 989 if is_fp16 else 1979
        if device_name.find("H200") >= 0:
            return 1979 if is_fp16 else 3958
        raise Exception(f"not supported device {device_name}")

    flops = M * N * K * 2
    return flops / _get_flops() / 1e9


def torch_allclose(x, y, rtol, atol):
    if not torch.allclose(x, y, rtol=rtol, atol=atol):
        print(f"shape of x: {x.shape}")
        print(f"shape of y: {y.shape}")

        print("x:", file=sys.stderr)
        print(x, file=sys.stderr)
        print("y:", file=sys.stderr)
        print(y, file=sys.stderr)
        print("x-y", x - y, file=sys.stderr)
        diff_loc = torch.isclose(x, y, rtol=rtol, atol=atol) == False
        print("x diff:", file=sys.stderr)
        print(x[diff_loc], file=sys.stderr)
        print("y diff:", file=sys.stderr)
        print(y[diff_loc], file=sys.stderr)
        num_diff = torch.sum(diff_loc)
        diff_rate = num_diff / (y.shape[0] * y.shape[1])
        print(f"diff count: {num_diff} ({diff_rate*100:.3f}%), {list(y.shape)}", file=sys.stderr)
        max_diff = torch.max(torch.abs(x - y))
        rtol_abs = rtol * torch.min(torch.abs(y))
        print(f"diff max: {max_diff}, atol: {atol}, rtol_abs: {rtol_abs}", file=sys.stderr)
        diff_indices = (diff_loc == True).nonzero(as_tuple=False)
        print(f"diff locations:\n{diff_indices}", file=sys.stderr)
        print("--------------------------------------------------------------\n", file=sys.stderr)
        raise RuntimeError


def get_torch_prof_ctx(do_prof: bool):
    ctx = (
        torch.profiler.profile(
            activities=[torch.profiler.ProfilerActivity.CPU, torch.profiler.ProfilerActivity.CUDA],
            record_shapes=True,
        )
        if do_prof
        else nullcontext()
    )
    return ctx


def is_fp8_dtype(dtype: torch.dtype) -> bool:
    return dtype in _get_torch_fp8_dtypes()


@contextmanager
def with_torch_deterministic(mode: bool, warn_only: bool = True):
    old_mode = torch.are_deterministic_algorithms_enabled()
    torch.use_deterministic_algorithms(mode, warn_only=warn_only)
    yield
    torch.use_deterministic_algorithms(old_mode, warn_only=warn_only)


__all__ = [
    "is_fp8_dtype",
    "get_arch",
    "estimate_gemm_sol_time_ms",
    "torch_allclose",
    "get_torch_prof_ctx",
    "is_fp8_dtype",
    "with_torch_deterministic",
]



################################################################################
#
# Copyright 2023 ByteDance Ltd. and/or its affiliates. All rights reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
################################################################################

import argparse
from functools import partial
import logging
import sys
import torch
import numpy as np
import torch.distributed
import flux
from utils import TP_GROUP, init_seed, RANK, WORLD_SIZE, NNODES, LOCAL_RANK, LOCAL_WORLD_SIZE


print = partial(print, flush=True)


@torch.no_grad()
def run_torch(
    input: torch.Tensor,
    weight: torch.Tensor,
    bias: torch.Tensor,
):
    # All gather input tensors from all gpus
    full_input = torch.empty(
        (input.size(0) * LOCAL_WORLD_SIZE, input.size(1)),
        dtype=input.dtype,
        device=torch.cuda.current_device(),
        requires_grad=False,
    )
    torch.distributed.all_gather_into_tensor(full_input, input, group=TP_GROUP)
    output = torch.matmul(full_input, weight.t())
    if bias is not None:
        output += bias

    return output, full_input


@torch.no_grad()
def run_flux_with_op(
    ag_op: flux.AGKernel,
    input: torch.Tensor,
    weight: torch.Tensor,
    bias: torch.Tensor,
    local_copy: bool = False,
):
    ag_op.reset_signals()
    if local_copy:
        ag_op.copy_local(input)
    output = ag_op.forward(input, weight, bias=bias)
    gathered = ag_op.gather()
    return output, gathered


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("M", type=int)
    parser.add_argument("N", type=int)
    parser.add_argument("K", type=int)
    parser.add_argument(
        "--ring_mode",
        default=-1,
        type=int,
        help="ring mode. -1 for auto detect. 0 for false, 1 for true",
    )
    return parser.parse_args()


def run_test_with_args(
    M_max,
    local_N,
    K,
    dtype,
    transpose_weight: bool,
    local_copy: bool,
    has_bias: bool,
    ring_mode: flux.AgRingMode,
):
    def _run(local_m):
        # input: [M_max, K], weight: [N, K]
        input = (
            (-2 * torch.rand((local_m, K), dtype=dtype).cuda() + 1) / 100 * (TP_GROUP.rank() + 1)
        )
        weight = (
            (-2 * torch.rand((local_N, K), dtype=dtype).cuda() + 1) / 100 * (TP_GROUP.rank() + 1)
        )
        bias = (
            torch.rand((local_m * LOCAL_WORLD_SIZE, local_N), dtype=dtype).cuda()
            / 10
            * (TP_GROUP.rank() + 1)
            if has_bias
            else None
        )
        w = weight.t().contiguous() if transpose_weight else weight.contiguous()

        gt_result, gt_all_input = run_torch(input, weight, bias)
        flux_result, flux_all_input = run_flux_with_op(
            ag_op,
            input,
            w,
            bias,
            local_copy,
        )

        torch.distributed.barrier()
        flux.bitwise_check(gt_all_input, flux_all_input)
        flux.torch_allclose(gt_result, flux_result, atol=1e-02, rtol=1e-02)

    ag_op = flux.AGKernel(
        TP_GROUP,
        NNODES,
        M_max,
        local_N,
        K,
        dtype,
        transpose_weight=transpose_weight,
        local_copy=local_copy,
        ring_mode=ring_mode,
    )

    torch.distributed.barrier(TP_GROUP)
    local_m_max = M_max // LOCAL_WORLD_SIZE
    try:
        _run(local_m_max)
        print(f"run with local_m={local_m_max} done")
    except:
        logging.exception(f"run with local_m={local_m_max} failed")
    try:
        _run(local_m_max // 2)
        print(f"run with local_m={local_m_max // 2} done")
    except:
        logging.exception(f"run with local_m={local_m_max//2} failed")


DTYPE_MAP = {"bfloat16": torch.bfloat16, "float16": torch.float16}

if __name__ == "__main__":
    torch.cuda.set_device(LOCAL_RANK)
    init_seed()
    args = parse_args()

    flux.init_flux_shm(TP_GROUP)
    torch.cuda.synchronize()

    assert args.M % TP_GROUP.size() == 0
    assert args.N % TP_GROUP.size() == 0
    assert args.K % TP_GROUP.size() == 0
    local_N = args.N // TP_GROUP.size()

    for dtype in [torch.float16, torch.bfloat16]:
        for local_copy in [True, False]:
            for transpose_weight in [True, False]:
                for has_bias in [True, False]:
                    print(
                        f"dtype={dtype}, local_copy={local_copy}, transpose_weight={transpose_weight}, has_bias={has_bias}"
                    )
                    run_test_with_args(
                        args.M,
                        local_N,
                        args.K,
                        dtype,
                        transpose_weight,
                        local_copy,
                        has_bias,
                        flux.AgRingMode(args.ring_mode),
                    )



from enum import Enum
import torch
import torch.distributed as dist
from typing import Optional, List

def bsr_reduce(input: torch.Tensor, output: torch.Tensor, block_h: int, block_w: int): ...
def bitwise_check(A: torch.Tensor, B: torch.Tensor) -> bool: ...
def uniform_initialize(tensor: torch.Tensor, seed: int, min: float, max: float) -> bool: ...

class TuningRecord:
    pass

class AgRingMode(Enum):
    All2All = ...
    Ring1D = ...
    Ring2D = ...
    Auto = ...

def load_tuning_record(record: TuningRecord) -> None: ...

class ProfilingContext:
    def __init__(self, name: str): ...
    def get_code(self) -> str: ...
    def get_latest_prof_result(self) -> str: ...
    def get_all_prof_results(self) -> List[str]: ...
    def get_latest_record(self) -> TuningRecord: ...
    def get_all_records(self) -> List[TuningRecord]: ...

class DistEnvTP:
    def __init__(tp_group: dist.ProcessGroup, nnodes: int) -> None: ...

class DistEnvTPWithEP:
    def __init__(
        tp_group: dist.ProcessGroup, nnodes: int, ep_group: Optional[dist.ProcessGroup] = None
    ) -> None: ...

class MoeArguments:
    def __init__(
        max_ntokens: int,
        hidden: int,
        ffn_hidden: int,
        nexperts: int,
        topk: int,
        input_dtype=torch.dtype,
        output_dtype: Optional[torch.dtype] = None,
    ) -> None: ...

class GemmOnly:
    def __init__(
        self,
        input_dtype: torch.dtype,
        output_dtype: Optional[torch.dtype] = None,
        transpose_weight: bool = False,
        use_fp8_gemm: bool = False,
    ): ...
    def forward(
        self,
        input: torch.Tensor,
        weight: torch.Tensor,
        bias: Optional[torch.Tensor] = None,
        output_buf: Optional[torch.Tensor] = None,
        input_scale: Optional[torch.Tensor] = None,
        weight_scale: Optional[torch.Tensor] = None,
        output_scale: Optional[torch.Tensor] = None,
        fast_accum: bool = False,
    ) -> torch.Tensor: ...
    def profiling(
        self,
        input: torch.Tensor,
        weight: torch.Tensor,
        bias: Optional[torch.Tensor] = None,
        output_buf: Optional[torch.Tensor] = None,
        input_scale: Optional[torch.Tensor] = None,
        weight_scale: Optional[torch.Tensor] = None,
        output_scale: Optional[torch.Tensor] = None,
        fast_accum: bool = False,
        prof_ctx: Optional[ProfilingContext] = None,
    ) -> torch.Tensor: ...

class GemmRS:
    def __init__(
        self,
        tp_group: dist.ProcessGroup,
        nnodes: int,
        max_m: int,
        n_dim: int,
        input_dtype: torch.dtype,
        output_dtype: Optional[torch.dtype] = None,
        transpose_weight=False,
        fuse_reduction=False,
    ): ...
    def forward(
        self,
        input: torch.Tensor,
        weight: torch.Tensor,
        bias: Optional[torch.Tensor] = None,
        input_scale: Optional[torch.Tensor] = None,
        weight_scale: Optional[torch.Tensor] = None,
        output_scale: Optional[torch.Tensor] = None,
        fast_accum: bool = False,
    ) -> torch.Tensor: ...
    def forward_gemm(
        self,
        input: torch.Tensor,
        weight: torch.Tensor,
        bias: Optional[torch.Tensor] = None,
        input_scale: Optional[torch.Tensor] = None,
        weight_scale: Optional[torch.Tensor] = None,
        output_scale: Optional[torch.Tensor] = None,
        fast_accum: bool = False,
    ) -> None: ...
    def forward_barrier(
        self, input: torch.Tensor, weight: torch.Tensor, bias: Optional[torch.Tensor] = None
    ) -> None: ...
    def forward_reduce_scatter(
        self, input: torch.Tensor, weight: torch.Tensor, bias: Optional[torch.Tensor] = None
    ) -> torch.Tensor: ...
    def zero_buffers(self) -> None: ...
    def profiling(
        self,
        input: torch.Tensor,
        weight: torch.Tensor,
        bias: Optional[torch.Tensor] = None,
        input_scale: Optional[torch.Tensor] = None,
        weight_scale: Optional[torch.Tensor] = None,
        output_scale: Optional[torch.Tensor] = None,
        fast_accum: bool = False,
        prof_ctx: Optional[ProfilingContext] = None,
    ) -> torch.Tensor: ...

class GemmRSAcrossNode:
    def __init__(
        self,
        rank: int,
        world_size: int,
        local_world_size: int,
        weight: torch.Tensor,
        transpose_weight: bool,
        output_buffer: torch.Tensor,
        output: torch.Tensor,
        barrier: torch.Tensor,
        fuse_reduction: bool,
    ): ...
    def forward(self, input: torch.Tensor) -> torch.Tensor: ...

class AllGatherGemm:
    def __init__(
        self,
        rank: int,
        world_size: int,
        input_buffer: torch.Tensor,
        weight: torch.Tensor,
        output_buffer: torch.Tensor,
        cuda_shm_handles: List[torch.Tensor],
        signal: torch.Tensor,
        signal_handles: List[torch.Tensor],
        signal_offsets: List[int],
        transpose_weight: bool,
        gather_output: bool,
    ): ...
    def forward(self, input: torch.Tensor) -> torch.Tensor: ...

class AGKernel:
    def __init__(
        self,
        tp_group: dist.ProcessGroup,
        nnodes: int,
        full_m: int,
        n_dim: int,
        k_dim: int,
        input_dtype: torch.dtype,
        output_dtype: Optional[torch.dtype] = None,
        transpose_weight: bool = True,
        local_copy: bool = True,
        ring_mode: AgRingMode = AgRingMode.Auto,
    ): ...
    def forward_allgather(self, input: torch.Tensor) -> None: ...
    def forward_gemm(
        self,
        input: torch.Tensor,
        weight: torch.Tensor,
        bias: Optional[torch.Tensor] = None,
        input_scale: Optional[torch.Tensor] = None,
        weight_scale: Optional[torch.Tensor] = None,
        output_scale: Optional[torch.Tensor] = None,
        fast_accum: bool = False,
    ) -> torch.Tensor: ...
    def forward(
        self,
        input: torch.Tensor,
        weight: torch.Tensor,
        bias: Optional[torch.Tensor] = None,
        input_scale: Optional[torch.Tensor] = None,
        weight_scale: Optional[torch.Tensor] = None,
        output_scale: Optional[torch.Tensor] = None,
        fast_accum: bool = False,
    ) -> torch.Tensor: ...
    def gather(self) -> torch.Tensor: ...
    def reset_signals(self) -> None: ...
    def copy_local(self, input: torch.Tensor) -> None: ...
    def profiling(
        self,
        input: torch.Tensor,
        weight: torch.Tensor,
        bias: Optional[torch.Tensor] = None,
        input_scale: Optional[torch.Tensor] = None,
        weight_scale: Optional[torch.Tensor] = None,
        output_scale: Optional[torch.Tensor] = None,
        fast_accum: bool = False,
        prof_ctx: Optional[ProfilingContext] = None,
    ) -> torch.Tensor: ...

class AGKernelCrossNode:
    def __init__(
        self,
        tp_group: dist.ProcessGroup,
        nnodes: int,
        output_buffer: torch.Tensor,
        full_m: int,
        n_dim: int,
        k_dim: int,
        input_dtype: torch.dtype,
        transpose_weight: bool = True,
        local_copy: bool = False,
        ring_mode: AgRingMode = AgRingMode.Auto,
    ): ...
    def reset_signals(self) -> None: ...
    def copy_local(self, input: torch.Tensor) -> None: ...
    def forward(self, input: torch.Tensor, weight: torch.Tensor) -> torch.Tensor: ...
    def gemm_only(
        self, input: torch.Tensor, full_input: torch.Tensor, weight: torch.Tensor
    ) -> torch.Tensor: ...

class All2AllOp:
    def __init__(self, rank: int, world_size: int, recv_buffer: torch.Tensor): ...
    def forward(self, send_buffer: List[torch.Tensor]) -> None: ...

class GemmGrouped:
    def __init__(self, weight: torch.Tensor, num_experts: int): ...
    def forward(self, input: torch.Tensor, split_cpu: torch.Tensor) -> torch.Tensor: ...

class GemmGroupedV3:
    def __init__(self, weight: torch.Tensor, num_experts: int): ...
    def forward(self, input: torch.Tensor, split_cpu: torch.Tensor) -> torch.Tensor: ...

class GemmGroupedV3AGScatter:
    def __init__(self, tp_env: DistEnvTP, moe_args: MoeArguments): ...
    def forward(
        self,
        inputs_shard: torch.Tensor,
        weights: torch.Tensor,
        splits_gpu: torch.Tensor,
        scatter_index: torch.Tensor,
        output_scale: Optional[torch.Tensor] = None,
        outputs_buf: Optional[torch.Tensor] = None,
        allgather_output: Optional[torch.Tensor] = None,
        fast_accum: bool = False,
        sm_mrgin: int = 0,
    ) -> torch.Tensor: ...
    def forward_multiple_weights(
        self,
        inputs_shard: torch.Tensor,
        weights: List[torch.Tensor],
        splits_gpu: torch.Tensor,
        scatter_index: torch.Tensor,
        output_scale: Optional[List[torch.Tensor]] = None,
        outputs_buf: Optional[List[torch.Tensor]] = None,
        allgather_output: Optional[torch.Tensor] = None,
        fast_accum: bool = False,
        sm_mrgin: int = 0,
    ) -> List[torch.Tensor]: ...
    def clear_buffers(self) -> None: ...
    def profiling(
        self,
        inputs_shard: torch.Tensor,
        weights: List[torch.Tensor],
        splits_gpu: torch.Tensor,
        scatter_index: torch.Tensor,
        output_scale: Optional[List[torch.Tensor]] = None,
        outputs_buf: Optional[List[torch.Tensor]] = None,
        allgather_output: Optional[torch.Tensor] = None,
        fast_accum: bool = False,
        sm_mrgin: int = 0,
        prof_ctx: Optional[ProfilingContext] = None,
    ) -> List[torch.Tensor]: ...

class GemmGroupedV3GatherRS:
    def __init__(
        self,
        num_experts: int,
        max_m: int,
        n_dim: int,
        topk: int,
        rank: int,
        world_size: int,
        tp_world_size: int,
        ep_world_size: int,
    ): ...
    def forward_gather_rs(
        self,
        input: torch.Tensor,
        weight: torch.Tensor,
        split_cpu: torch.Tensor,
        scatter_idx: torch.Tensor,
        input_scale: Optional[torch.Tensor] = None,
        weight_scale: Optional[torch.Tensor] = None,
        output_scale: Optional[torch.Tensor] = None,
        fastacc: bool = True,
        sm_mrgin: int = 0,
        with_stream_sync: bool = False,
    ) -> torch.Tensor: ...
    def forward_gather_rs_multiple(
        self,
        input: List[torch.Tensor],
        weight: List[torch.Tensor],
        split_cpu: torch.Tensor,
        scatter_idx: torch.Tensor,
        input_scale: Optional[List[torch.Tensor]] = None,
        weight_scale: Optional[List[torch.Tensor]] = None,
        output_scale: Optional[List[torch.Tensor]] = None,
        fastacc: bool = True,
        sm_mrgin: int = 0,
        with_stream_sync: bool = False,
    ) -> torch.Tensor: ...



################################################################################
#
# Copyright 2023 ByteDance Ltd. and/or its affiliates. All rights reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
################################################################################

# usage: torchrun --node_rank=0 --nproc_per_node=8 --nnodes=1 --rdzv_id=none --master_addr=127.0.0.1 --master_port=23456 test/test_ag_kernel.py 4096 49152 12288
import argparse
from functools import partial
import os
import time
import torch
import numpy as np
import datetime
import torch.distributed
from contextlib import nullcontext
import flux

RANK = int(os.environ.get("RANK", 0))
LOCAL_RANK = int(os.environ.get("LOCAL_RANK", 0))
WORLD_SIZE = int(os.environ.get("WORLD_SIZE", 1))
LOCAL_WORLD_SIZE = int(os.environ.get("LOCAL_WORLD_SIZE", 1))
NNODES = WORLD_SIZE // LOCAL_WORLD_SIZE

torch.cuda.set_device(LOCAL_RANK)
os.environ["NCCL_DEBUG"] = os.getenv("NCCL_DEBUG", "ERROR")
os.environ["CUBLAS_WORKSPACE_CONFIG"] = ":16:8"
torch.use_deterministic_algorithms(True, warn_only=True)
torch.set_printoptions(precision=2)
torch.manual_seed(3 + RANK)
torch.cuda.manual_seed_all(3 + RANK)
torch.backends.cudnn.deterministic = True
torch.backends.cudnn.benchmark = False
torch.backends.cuda.matmul.allow_tf32 = False
np.random.seed(3 + RANK)
# torch.set_printoptions(profile="full") ## debug

torch.distributed.init_process_group(
    backend="nccl",
    world_size=WORLD_SIZE,
    rank=RANK,
    timeout=datetime.timedelta(seconds=1800),
)
# use all ranks as tp group
TP_GROUP = torch.distributed.new_group(ranks=list(range(WORLD_SIZE)), backend="nccl")
print = partial(print, flush=True)


class PerfResult:
    def __init__(
        self,
        name: str,
        output: torch.Tensor,
        total_ms: float,
        time1: str,
        gemm_time_ms: float,
        time2: str,
        comm_time_ms: float,
        time3: str = "gemm_only",
        gemm_only_time_ms: float = 0,
    ) -> None:
        self.name = name
        self.output = output
        self.total_ms = total_ms
        self.time1 = time1
        self.time2 = time2
        self.gemm_time_ms = gemm_time_ms
        self.comm_time_ms = comm_time_ms

        self.time3 = time3
        self.gemm_only_time_ms = gemm_only_time_ms

    def __repr__(self) -> str:
        return (
            f"{self.name}: total {self.total_ms:.3f} ms, {self.time1} {self.gemm_time_ms:.3f} ms"
            f", {self.time2} {self.comm_time_ms:.3f} ms, {self.time3} {self.gemm_only_time_ms:.3f} ms"
        )


@torch.no_grad()
def perf_torch(input: torch.Tensor, weight: torch.Tensor, warmup: int, iters: int):
    local_M = input.size(0)
    M = local_M * TP_GROUP.size()

    torch.distributed.barrier()
    # All gather input tensors from all gpus
    full_input = torch.zeros(
        (M, input.size(1)),
        dtype=input.dtype,
        device=torch.cuda.current_device(),
        requires_grad=False,
    )
    torch.distributed.all_gather_into_tensor(full_input, input, group=TP_GROUP)

    torch.distributed.barrier()
    warmup_iters = warmup
    total_iters = warmup_iters + iters
    start_events = [torch.cuda.Event(enable_timing=True) for _ in range(total_iters)]
    allgather_end_events = [torch.cuda.Event(enable_timing=True) for _ in range(total_iters)]
    end_events = [torch.cuda.Event(enable_timing=True) for _ in range(total_iters)]

    torch.distributed.barrier()
    for i in range(total_iters):
        start_events[i].record()
        torch.distributed.all_gather_into_tensor(full_input, input, group=TP_GROUP)
        allgather_end_events[i].record()
        output = torch.matmul(full_input, weight.t())
        end_events[i].record()

    comm_times = []  # all gather
    gemm_times = []  # gemm
    for i in range(total_iters):
        allgather_end_events[i].synchronize()
        end_events[i].synchronize()
        if i >= warmup_iters:
            comm_times.append(start_events[i].elapsed_time(allgather_end_events[i]) / 1000)
            gemm_times.append(allgather_end_events[i].elapsed_time(end_events[i]) / 1000)

    comm_time = sum(comm_times) / iters * 1000
    gemm_time = sum(gemm_times) / iters * 1000

    return PerfResult(
        name=f"torch #{TP_GROUP.rank()}",
        output=output,
        total_ms=gemm_time + comm_time,
        time1="gemm",
        gemm_time_ms=gemm_time,
        time2="comm",
        comm_time_ms=comm_time,
    )


@torch.no_grad()
def perf_flux(
    input: torch.Tensor,
    weight: torch.Tensor,
    warmup: int,
    iters: int,
    transpose_weight: bool = True,
    local_copy: bool = False,
    gather_output: bool = False,
    ring_mode: flux.AgRingMode = flux.AgRingMode.Auto,
):
    local_M = input.size(0)
    M = local_M * TP_GROUP.size()
    K = input.size(1)

    if transpose_weight:
        w = weight.t().contiguous()
        N = w.size(1)
    else:
        w = weight
        N = w.size(0)

    torch.distributed.barrier()
    full_input = torch.zeros(
        (M, K),
        dtype=input.dtype,
        device=torch.cuda.current_device(),
        requires_grad=False,
    )
    torch.distributed.all_gather_into_tensor(full_input, input, group=TP_GROUP)
    torch.distributed.barrier()

    gemm_only_op = flux.GemmOnly(w.dtype, transpose_weight=transpose_weight)
    all_gather_gemm_kernel = flux.AGKernelXNode(
        TP_GROUP,
        NNODES,
        M,
        N,
        K,
        input.dtype,
        transpose_weight,
        gather_output,
        local_copy,
        ring_mode=ring_mode,
    )

    warmup_iters = warmup
    total_iters = warmup_iters + iters
    start_events = [torch.cuda.Event(enable_timing=True) for _ in range(total_iters)]
    end_events = [torch.cuda.Event(enable_timing=True) for _ in range(total_iters)]
    gemm_only_output_buf = torch.empty(
        [M, N], dtype=input.dtype, device=input.device, requires_grad=False
    )

    torch.distributed.barrier()
    for i in range(total_iters):
        start_events[i].record()
        _ = gemm_only_op.forward(full_input, w, output_buf=gemm_only_output_buf)
        end_events[i].record()
    torch.cuda.current_stream().synchronize()

    gemm_times = []
    for i in range(total_iters):
        end_events[i].synchronize()
        if i >= warmup_iters:
            gemm_times.append(start_events[i].elapsed_time(end_events[i]) / 1000)
    gemm_time = sum(gemm_times)

    time.sleep(1)

    torch.distributed.barrier()
    for i in range(total_iters):
        all_gather_gemm_kernel.reset_signals()
        if local_copy:
            all_gather_gemm_kernel.copy_local(input)
        start_events[i].record()
        output, gathered = all_gather_gemm_kernel.forward(input, w)
        end_events[i].record()

    torch.distributed.barrier()
    torch.cuda.current_stream().synchronize()

    ag_gemm_times = []
    for i in range(total_iters):
        end_events[i].synchronize()
        if i >= warmup_iters:
            ag_gemm_times.append(start_events[i].elapsed_time(end_events[i]) / 1000)
    ag_gemm_time = sum(ag_gemm_times)

    time.sleep(1)

    torch.distributed.barrier()

    ## signals are already set
    for i in range(total_iters):
        start_events[i].record()
        _ = all_gather_gemm_kernel.gemm_only(input, full_input, w)
        end_events[i].record()

    torch.distributed.barrier()
    torch.cuda.current_stream().synchronize()

    gemm_only_times = []
    for i in range(total_iters):
        end_events[i].synchronize()
        if i >= warmup_iters:
            gemm_only_times.append(start_events[i].elapsed_time(end_events[i]) / 1000)

    gemm_only_time = sum(gemm_only_times)

    gemm_only_time_ms = gemm_only_time / iters * 1000

    ag_gemm_time_ms = ag_gemm_time / iters * 1000
    gemm_time_ms = gemm_time / iters * 1000
    comm_time_ms = (ag_gemm_time - gemm_time) / iters * 1000

    return PerfResult(
        name=f"flux  #{TP_GROUP.rank()}",
        output=output,
        total_ms=ag_gemm_time_ms,
        time1="gemm",
        gemm_time_ms=gemm_time_ms,
        time2="comm",
        comm_time_ms=comm_time_ms,
        time3="gemm_only",
        gemm_only_time_ms=gemm_only_time_ms,
    )


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("M", type=int)
    parser.add_argument("N", type=int)
    parser.add_argument("K", type=int)
    parser.add_argument("--warmup", default=5, type=int, help="warmup iterations")
    parser.add_argument("--iters", default=10, type=int, help="perf iterations")
    parser.add_argument("--dtype", default="bfloat16", type=str, help="data type")
    parser.add_argument(
        "--profile",
        default=False,
        action="store_true",
        help="dump torch.profiler.profile",
    )
    parser.add_argument(
        "--local_copy",
        default=True,
        action=argparse.BooleanOptionalAction,
        help="perform local copy",
    )
    parser.add_argument(
        "--gather_output",
        default=False,
        action="store_true",
        help="output gather results",
    )
    parser.add_argument(
        "--transpose_weight",
        action=argparse.BooleanOptionalAction,
        default=True,
        help="transpose weight",
    )
    parser.add_argument(
        "--ring_mode",
        default=-1,
        type=int,
        help="ring mode. -1 for auto detect. 0 for all-to-all, 1 for 1d ring. 2 for 2d ring. 3 for custom ring.",
    )
    return parser.parse_args()


DTYPE_MAP = {"bfloat16": torch.bfloat16, "float16": torch.float16}

if __name__ == "__main__":
    torch.cuda.set_device(LOCAL_RANK)
    args = parse_args()
    print("before flux_shm initialization")
    flux.init_flux_shm(TP_GROUP)
    torch.cuda.synchronize()
    print("after flux_shm initialization")

    dtype = DTYPE_MAP[args.dtype]
    assert args.M % TP_GROUP.size() == 0
    assert args.N % TP_GROUP.size() == 0
    assert args.K % TP_GROUP.size() == 0
    local_M = args.M // TP_GROUP.size()
    local_N = args.N // TP_GROUP.size()

    # input: [M, K], weight: [N, K]
    input = (
        (-2 * torch.rand((local_M, args.K), dtype=dtype).cuda() + 1) / 100 * (TP_GROUP.rank() + 1)
    )
    weight = (
        (-2 * torch.rand((local_N, args.K), dtype=dtype).cuda() + 1) / 100 * (TP_GROUP.rank() + 1)
    )

    ctx = (
        torch.profiler.profile(
            activities=[
                torch.profiler.ProfilerActivity.CPU,
                torch.profiler.ProfilerActivity.CUDA,
            ],
            record_shapes=True,
            with_stack=True,
        )
        if args.profile
        else nullcontext()
    )

    with ctx:
        perf_res_torch = perf_torch(input, weight, args.warmup, args.iters)
        perf_res_flux = perf_flux(
            input,
            weight,
            args.warmup,
            args.iters,
            args.transpose_weight,
            args.local_copy,
            args.gather_output,
            flux.AgRingMode(args.ring_mode),
        )

    if args.profile:
        run_id = os.environ["TORCHELASTIC_RUN_ID"]
        import pathlib
        import sys

        prof_dir = f"prof/{pathlib.Path(sys.argv[0]).stem}_{run_id}"
        os.makedirs(prof_dir, exist_ok=True)
        ctx.export_chrome_trace(f"{prof_dir}/trace_rank{TP_GROUP.rank()}.json.gz")

    if TP_GROUP.rank() == 0:
        print(
            f"SOL time for GEMM(M={args.M},N={args.N},K={args.K},TP={TP_GROUP.size()}):"
            f" {flux.estimate_gemm_sol_time_ms(local_M, args.N, args.K):.3f}ms"
        )

    for i in range(TP_GROUP.size()):
        if i == TP_GROUP.rank():
            print(perf_res_torch)
        torch.distributed.barrier()
    torch.distributed.barrier()
    for i in range(TP_GROUP.size()):
        if i == TP_GROUP.rank():
            print(perf_res_flux)
        torch.distributed.barrier()
    torch.distributed.barrier()

    flux_output = perf_res_flux.output
    torch_output = perf_res_torch.output
    flux.torch_allclose(flux_output, torch_output, atol=1e-02, rtol=1e-02)



################################################################################
#
# Copyright 2023 ByteDance Ltd. and/or its affiliates. All rights reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
################################################################################

import argparse
import os
import time
import datetime
import torch
import numpy as np
import flux
from flux.util import is_fp8_dtype

os.environ["CUBLAS_WORKSPACE_CONFIG"] = ":16:8"
torch.use_deterministic_algorithms(True, warn_only=True)
torch.set_printoptions(precision=8)
torch.manual_seed(3)
torch.cuda.manual_seed_all(3)
torch.backends.cudnn.deterministic = True
torch.backends.cudnn.benchmark = False
torch.backends.cuda.matmul.allow_tf32 = False
torch.backends.cuda.matmul.allow_fp16_reduced_precision_reduction = False
torch.backends.cuda.matmul.allow_bf16_reduced_precision_reduction = False
np.random.seed(3)


class PerfResult:
    def __init__(self, name: str, output: torch.Tensor, gemm_time_ms: float) -> None:
        self.name = name
        self.output = output
        self.gemm_time_ms = gemm_time_ms

    def __repr__(self) -> str:
        return f"{self.name}: gemm {self.gemm_time_ms:.3f} ms"


def perf_gemm(iters: int, name: str, fn: callable):
    warmup_iters = 5
    for i in range(warmup_iters):
        output = fn()
    torch.cuda.synchronize()
    total_time = 0
    start = time.time()
    for i in range(iters):
        output = fn()
    torch.cuda.synchronize()
    end = time.time()
    total_time = end - start
    return PerfResult(name=name, output=output, gemm_time_ms=total_time / iters * 1000)


def perf_torch(
    input: torch.Tensor,
    weight: torch.Tensor,
    bias: torch.Tensor,
    input_scale: torch.Tensor,
    weight_scale: torch.Tensor,
    is_fp8: bool,
    iters: int,
):

    alpha_scale = 1.0
    if is_fp8:
        alpha_scale = input_scale * weight_scale
        input = input.to(torch.bfloat16)
        weight = weight.to(torch.bfloat16)

    def fn():
        output = alpha_scale * torch.matmul(input, weight.t())
        if is_fp8:
            output = output.to(torch.bfloat16)
        if bias is not None:
            output = output + bias
        return output

    return perf_gemm(iters, "torch", fn)


def perf_flux(
    input: torch.Tensor,
    weight: torch.Tensor,
    bias: torch.Tensor,
    input_scale: torch.Tensor,
    weight_scale: torch.Tensor,
    transpose_weight: bool,
    is_fp8: bool,
    iters: int,
):
    m = input.size(0)
    if transpose_weight:
        assert is_fp8 == False, "FP8 GEMM does not support RRR layout"
        weight = weight.t().contiguous()
        n = weight.size(1)
    else:
        n = weight.size(0)

    output_dtype = torch.bfloat16 if is_fp8 else input.dtype
    output = torch.empty([m, n], dtype=output_dtype, device=input.device, requires_grad=False)
    ## TODO: remove below once moe fp8 gemm invoke get fixed
    use_fp8_gemm = True if is_fp8 else False
    op = flux.GemmOnly(
        input_dtype=input.dtype,
        output_dtype=output_dtype,
        transpose_weight=transpose_weight,
        use_fp8_gemm=use_fp8_gemm,
    )

    def fn():
        return op.forward(
            input,
            weight,
            bias=bias,
            output_buf=output,
            input_scale=input_scale,
            weight_scale=weight_scale,
            output_scale=None,
            fast_accum=False,
        )

    return perf_gemm(iters, "flux", fn)


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("M", type=int)
    parser.add_argument("N", type=int)
    parser.add_argument("K", type=int)
    parser.add_argument("--iters", default=50, type=int, help="perf iterations")
    parser.add_argument(
        "--dtype",
        default="bfloat16",
        type=str,
        help="allowed data type:: bfloat16(default),float16,fp8e4m3,fp8e5m2.",
    )
    parser.add_argument(
        "--has_bias", default=False, action="store_true", help="whether to add bias"
    )
    parser.add_argument(
        "--transpose_weight", default=False, action="store_true", help="whether to transpose weight"
    )

    return parser.parse_args()


DTYPE_MAP = {
    "bfloat16": torch.bfloat16,
    "float16": torch.float16,
    "fp8e4m3": torch.float8_e4m3fn,
    "fp8e5m2": torch.float8_e5m2,
}

THRESHOLD_MAP = {
    torch.float16: 1e-2,
    torch.bfloat16: 2e-2,
    torch.float8_e4m3fn: 2e-2,
    torch.float8_e5m2: 2e-2,
}

if __name__ == "__main__":
    args = parse_args()
    dtype = DTYPE_MAP[args.dtype]
    is_fp8 = is_fp8_dtype(dtype)

    input = None
    weight = None
    if is_fp8:
        torch.use_deterministic_algorithms(False, warn_only=True)
        input = torch.rand((args.M, args.K), dtype=torch.bfloat16).cuda() / 10
        weight = torch.rand((args.N, args.K), dtype=torch.bfloat16).cuda() / 10
        # convert to fp8
        input = input.to(dtype)
        weight = weight.to(dtype)
    else:
        input = torch.rand((args.M, args.K), dtype=dtype).cuda() / 10
        weight = torch.rand((args.N, args.K), dtype=dtype).cuda() / 10

    input_scale = None
    weight_scale = None

    if is_fp8:
        input_scale = torch.rand(1, dtype=torch.float32).cuda()
        weight_scale = torch.rand(1, dtype=torch.float32).cuda()

    bias = None
    if is_fp8:
        bias = torch.rand((1, args.N), dtype=torch.bfloat16).cuda() if args.has_bias else None
    else:
        bias = torch.rand((args.M, args.N), dtype=dtype).cuda() if args.has_bias else None

    perf_result_flux = perf_flux(
        input, weight, bias, input_scale, weight_scale, args.transpose_weight, is_fp8, args.iters
    )
    perf_result_torch = perf_torch(
        input, weight, bias, input_scale, weight_scale, is_fp8, args.iters
    )

    print(f"SOL gemm {flux.estimate_gemm_sol_time_ms(args.M, args.N, args.K):.3f}ms")
    print(perf_result_torch)
    print(perf_result_flux)

    flux_output = perf_result_flux.output
    torch_output = perf_result_torch.output
    atol = THRESHOLD_MAP[dtype]
    rtol = THRESHOLD_MAP[dtype]
    flux.torch_allclose(flux_output, torch_output, atol=atol, rtol=rtol)



################################################################################
#
# Copyright 2023 ByteDance Ltd. and/or its affiliates. All rights reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
################################################################################

from contextlib import contextmanager
import ctypes
import os
import sys
from pathlib import Path
import torch
import importlib
import logging

FLUX_TORCH_EXTENSION_NAME = "flux_ths_pybind"


def _preload_libs(libname):
    libpath = Path(__file__).parent.parent / "lib" / libname
    try:
        ctypes.CDLL(libpath)
    except OSError as e:
        # Try to load from the LD_LIBRARY_PATH
        ctypes.CDLL(libname)


def _load_deps():
    _preload_libs("libflux_cuda.so")
    try:
        _preload_libs("nvshmem_bootstrap_torch.so")
        _preload_libs("nvshmem_transport_ibrc.so.2")
    except Exception as e:
        logging.info("Failed to load NVSHMEM libs")


_load_deps()
flux_mod = importlib.import_module(FLUX_TORCH_EXTENSION_NAME)


class NotCompiled:
    pass


bsr_reduce = getattr(flux_mod, "bsr_reduce", NotCompiled())
init_flux_shm = flux_mod.init_flux_shm
nvshmem_create_tensor = getattr(flux_mod, "nvshmem_create_tensor", NotCompiled())
bitwise_check = flux_mod.bitwise_check
uniform_initialize = flux_mod.uniform_initialize
load_tuning_record = flux_mod.load_tuning_record

ProfilingContext = flux_mod.ProfilingContext
TuningRecord = flux_mod.TuningRecord
DistEnvTP = flux_mod.DistEnvTP
DistEnvTPWithEP = flux_mod.DistEnvTPWithEP
MoeArguments = flux_mod.MoeArguments
GemmOnly = getattr(flux_mod, "GemmOnly", NotCompiled())
GemmRS = getattr(flux_mod, "GemmRS", NotCompiled())
GemmRSAcrossNode = getattr(flux_mod, "GemmRSAcrossNode", NotCompiled())
AllGatherGemm = getattr(flux_mod, "AllGatherGemm", NotCompiled())
AGKernel = getattr(flux_mod, "AGKernel", NotCompiled())
AGKernelCrossNode = getattr(flux_mod, "AGKernelCrossNode", NotCompiled())
AGKernelCrossNodeNvshmem = getattr(flux_mod, "AGKernelCrossNodeNvshmem", NotCompiled())
GemmGroupedV3AGScatter = getattr(flux_mod, "GemmGroupedV3AGScatter", NotCompiled())
GemmGroupedV3GatherRS = getattr(flux_mod, "GemmGroupedV3GatherRS", NotCompiled())
AgRingMode = getattr(flux_mod, "AGRingMode", NotCompiled())


def get_from_torch_classes(name: str):
    try:
        return getattr(torch.classes.flux, name)
    except:
        return NotCompiled()


All2AllOp = get_from_torch_classes("All2AllOp")
GemmGrouped = get_from_torch_classes("GemmGrouped")
GemmGroupedV3 = get_from_torch_classes("GemmGroupedV3")
# GemmGroupedV3GatherRS = get_from_torch_classes("GemmGroupedV3GatherRS")


__all__ = [
    "bsr_reduce",
    "init_flux_shm",
    "bitwise_check",
    "uniform_initialize",
    "load_tuning_record",
    "TuningRecord",
    "ProfilingContext",
    "DistEnvTP",
    "DistEnvTPWithEP",
    "MoeArguments",
    "GemmOnly",
    "GemmRS",
    "GemmRSAcrossNode",
    "AllGatherGemm",
    "AGKernel",
    "AGKernelCrossNode",
    "All2AllOp",
    "GemmGrouped",
    "GemmGroupedV3",
    "GemmGroupedV3AGScatter",
    "GemmGroupedV3GatherRS",
    "AgRingMode",
    "nvshmem_create_tensor",
]



################################################################################
#
# Copyright 2023 ByteDance Ltd. and/or its affiliates. All rights reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
################################################################################

import torch
import torch.distributed as dist
from . import cpp_mod

_INTER_NODE_AG_GEMM_GROUP = None
_INTRA_NODE_AG_GEMM_GROUP = None


def get_inter_node_ag_gemm_group(tp_group: dist.ProcessGroup, nnodes: int):
    global _INTER_NODE_AG_GEMM_GROUP
    if _INTER_NODE_AG_GEMM_GROUP is not None:
        return _INTER_NODE_AG_GEMM_GROUP

    tp_world_size: int = tp_group.size()
    assert tp_world_size % nnodes == 0, f"{tp_world_size} is not divisible by {nnodes}"
    local_world_size: int = tp_group.size() // nnodes
    world_size: int = dist.get_world_size()
    assert world_size % tp_world_size == 0, f"{world_size} not divisible by {tp_world_size}"

    ranks_per_sub_group = []
    for tp0_rank in range(0, world_size, tp_world_size):
        for local_rank in range(local_world_size):
            inter_ranks = [
                tp0_rank + node_rank * local_world_size + local_rank for node_rank in range(nnodes)
            ]
            ranks_per_sub_group.append(inter_ranks)
    _INTER_NODE_AG_GEMM_GROUP, _ = dist.new_subgroups_by_enumeration(ranks_per_sub_group)
    return _INTER_NODE_AG_GEMM_GROUP


def get_intra_node_ag_gemm_group(tp_group: dist.ProcessGroup, nnodes: int):
    global _INTRA_NODE_AG_GEMM_GROUP
    if _INTRA_NODE_AG_GEMM_GROUP is not None:
        return _INTRA_NODE_AG_GEMM_GROUP

    tp_world_size: int = tp_group.size()
    assert tp_world_size % nnodes == 0, f"{tp_world_size} is not divisible by {nnodes}"
    local_world_size: int = tp_group.size() // nnodes
    world_size: int = dist.get_world_size()
    assert world_size % tp_world_size == 0, f"{world_size} not divisible by {tp_world_size}"

    ranks_per_sub_group = []
    for node_id in range(nnodes):
        ranks_per_sub_group.append(
            list(range(node_id * local_world_size, (node_id + 1) * local_world_size))
        )
    _INTRA_NODE_AG_GEMM_GROUP, _ = dist.new_subgroups_by_enumeration(ranks_per_sub_group)
    return _INTRA_NODE_AG_GEMM_GROUP


class AGGemm0:
    def __init__(
        self,
        weight: torch.Tensor,
        tp_group,
        full_m: int,
        nnodes: int = 1,
        transpose_weight: bool = False,
        gather_output=False,
    ) -> None:
        self.weight = weight
        self.tp_group = tp_group
        self.rank = self.tp_group.rank()
        self.world_size = self.tp_group.size()
        self.full_m = full_m
        self.nnodes = nnodes
        assert self.world_size % nnodes == 0
        assert self.full_m % self.world_size == 0
        self.local_world_size = self.world_size // nnodes
        self.local_rank = self.rank % self.local_world_size
        if nnodes > 1:
            self.inter_node_group = get_inter_node_ag_gemm_group(tp_group, nnodes)
        self.intra_node_group = get_intra_node_ag_gemm_group(tp_group, nnodes)

        self.transpose_weight = transpose_weight
        self.gather_output = gather_output
        self.n_dim = weight.size(1) if transpose_weight else weight.size(0)
        self.k_dim = weight.size(0) if transpose_weight else weight.size(1)

        # input buffer to hold data from other devices
        self.input_buffer = None
        self.shm_handles = None
        self.shm_offsets = None

        self.output_buffer = None

        # signal for dependency
        self.signal = None
        self.signal_handles = None
        self.signal_offsets = None

        self.cpp_op = None
        self.lazy_init_cpp_op(self.full_m)

    def lazy_init_cpp_op(self, full_m: int):
        if self.cpp_op is not None:
            return

        self.full_m = full_m

        # input buffer and handle
        self.input_buffer = torch.zeros(
            (full_m, self.k_dim),
            dtype=self.weight.dtype,
            device=self.weight.device,
            requires_grad=False,
        )
        input_buffer_offset = self.input_buffer.storage_offset()
        shm_handle = self.input_buffer._typed_storage()._share_cuda_()[1]
        shm_offset = self.input_buffer._typed_storage()._share_cuda_()[3]
        shm_handle_ts_cuda = torch.ByteTensor(torch.ByteStorage._from_buffer(shm_handle)).cuda()
        shm_handles = [torch.empty_like(shm_handle_ts_cuda) for _ in range(self.local_world_size)]
        torch.distributed.all_gather(shm_handles, shm_handle_ts_cuda, group=self.intra_node_group)
        self.shm_handles = [handle.cpu() for handle in shm_handles]

        offset_value = shm_offset + input_buffer_offset
        offset_list = [None for _ in range(self.local_world_size)]
        torch.distributed.all_gather_object(offset_list, offset_value, group=self.intra_node_group)
        self.shm_offsets = offset_list

        # output buffer
        self.output_buffer = torch.zeros(
            (full_m, self.n_dim),
            dtype=self.weight.dtype,
            device=self.weight.device,
            requires_grad=False,
        )

        # cuda signal
        num_stages = self.local_world_size
        self.signal = torch.zeros(
            (num_stages),
            dtype=torch.int64,
            device=self.weight.device,
            requires_grad=False,
        )
        signal_tensor_offset = self.signal.storage_offset()

        signal_handle = self.signal._typed_storage()._share_cuda_()[1]
        signal_offset = self.signal._typed_storage()._share_cuda_()[3]
        signal_handle_ts_cuda = torch.ByteTensor(
            torch.ByteStorage._from_buffer(signal_handle)
        ).cuda()
        signal_handles = [
            torch.empty_like(signal_handle_ts_cuda) for _ in range(self.local_world_size)
        ]
        torch.distributed.all_gather(
            signal_handles, signal_handle_ts_cuda, group=self.intra_node_group
        )
        self.signal_handles = [handle.cpu() for handle in signal_handles]

        signal_total_offset = signal_tensor_offset + signal_offset
        signal_offset_list = [None for _ in range(self.local_world_size)]
        torch.distributed.all_gather_object(
            signal_offset_list, signal_total_offset, group=self.intra_node_group
        )
        self.signal_offsets = signal_offset_list

        self.cpp_op = cpp_mod.AllGatherGemm(
            self.local_rank,
            self.local_world_size,
            self.input_buffer,
            self.weight,
            self.output_buffer,
            self.shm_handles,
            self.shm_offsets,
            self.signal,
            self.signal_handles,
            self.signal_offsets,
            self.transpose_weight,
            self.gather_output,
        )

    def _lazy_init_inter_ag_output_buffer(self, input: torch.Tensor):
        input_shape = input.size()
        m_dim, k_dim = input_shape[-2], input_shape[-1]
        self.inter_ag_output_buffer = torch.empty(
            (m_dim * self.nnodes, k_dim),
            dtype=input.dtype,
            device=input.device,
            requires_grad=False,
        )

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        if self.nnodes > 1:
            self._lazy_init_inter_ag_output_buffer(input)
            dist.all_gather_into_tensor(
                self.inter_ag_output_buffer, input, group=self.inter_node_group
            )
            input = self.inter_ag_output_buffer

        self.lazy_init_cpp_op(self.full_m)
        full_output = self.cpp_op.forward(input)
        gather_out = self.input_buffer if self.gather_output else None
        return full_output, gather_out



################################################################################
#
# Copyright 2023 ByteDance Ltd. and/or its affiliates. All rights reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
################################################################################

# usage: torchrun --node_rank=0 --nproc_per_node=8 --nnodes=1 --rdzv_id=none --master_addr=127.0.0.1 --master_port=23456 test/test_gemm_rs.py 2048 10240 40960
import argparse
from functools import partial
import os
import sys
import time
from typing import Union
import torch
import numpy as np
import datetime
import torch.distributed
from contextlib import nullcontext
import flux
from flux.gemm_rs_sm80 import get_intra_node_pg_group

RANK = int(os.environ.get("RANK", 0))
LOCAL_RANK = int(os.environ.get("LOCAL_RANK", 0))
LOCAL_WORLD_SIZE = int(os.environ.get("LOCAL_WORLD_SIZE", 1))
WORLD_SIZE = int(os.environ.get("WORLD_SIZE", 1))
NNODES = WORLD_SIZE // LOCAL_WORLD_SIZE
torch.cuda.set_device(LOCAL_RANK)

os.environ["NCCL_DEBUG"] = "ERROR"
os.environ["CUBLAS_WORKSPACE_CONFIG"] = ":16:8"
torch.use_deterministic_algorithms(True, warn_only=True)
torch.set_printoptions(precision=8)
torch.manual_seed(3 + RANK)
torch.cuda.manual_seed_all(3 + RANK)
torch.backends.cudnn.deterministic = True
torch.backends.cudnn.benchmark = False
torch.backends.cuda.matmul.allow_tf32 = False
torch.backends.cuda.matmul.allow_fp16_reduced_precision_reduction = False
torch.backends.cuda.matmul.allow_bf16_reduced_precision_reduction = False
np.random.seed(3 + RANK)

torch.distributed.init_process_group(
    backend="nccl", world_size=WORLD_SIZE, rank=RANK, timeout=datetime.timedelta(seconds=1800)
)
# use all ranks as tp group
TP_GROUP = torch.distributed.new_group(ranks=list(range(WORLD_SIZE)), backend="nccl")
print = partial(print, flush=True)


class PerfResult:
    def __init__(
        self, name: str, output: torch.Tensor, gemm_time_ms: float, comm_time_ms: float
    ) -> None:
        self.name = name
        self.output = output
        self.gemm_time_ms = gemm_time_ms
        self.comm_time_ms = comm_time_ms

    def __repr__(self) -> str:
        return (
            f"{self.name}: gemm {self.gemm_time_ms:.3f} ms, comm {self.comm_time_ms:.3f} ms"
            f", total {self.gemm_time_ms + self.comm_time_ms:.3f} ms"
        )


@torch.no_grad()
def perf_torch(
    input: torch.Tensor,
    weight: torch.Tensor,
    bias: torch.Tensor,
    warmup: int,
    iters: int,
    transpose_weight: bool = False,
    input_scale: Union[None, torch.Tensor] = None,
    weight_scale: Union[None, torch.Tensor] = None,
):
    torch.distributed.barrier()

    is_fp8 = flux.util.is_fp8_dtype(input.dtype)
    warmup_iters = warmup
    output_dtype = input.dtype if not is_fp8 else torch.bfloat16
    m = input.size(0)
    with flux.util.with_torch_deterministic(False):
        if transpose_weight:
            w = weight.t().contiguous()
            n = w.size(1)
        else:
            n = weight.size(0)
            w = weight

        full_output = torch.zeros(
            [m, n],
            dtype=output_dtype,
            device=torch.cuda.current_device(),
            requires_grad=False,
        )
        output = torch.zeros(
            [m // WORLD_SIZE, n],
            dtype=output_dtype,
            device=torch.cuda.current_device(),
            requires_grad=False,
        )

    op = (
        flux.GemmOnly(
            input_dtype=input.dtype,
            output_dtype=output_dtype,
            transpose_weight=transpose_weight,
            use_fp8_gemm=is_fp8,
        )
        if is_fp8
        else None
    )

    total_iters = warmup_iters + iters
    start_events = [torch.cuda.Event(enable_timing=True) for _ in range(total_iters)]
    gemm_end_events = [torch.cuda.Event(enable_timing=True) for _ in range(total_iters)]
    end_events = [torch.cuda.Event(enable_timing=True) for _ in range(total_iters)]
    torch.distributed.barrier()
    for i in range(total_iters):
        start_events[i].record()
        if is_fp8:
            full_output = op.forward(
                input,
                w,
                bias=bias,
                output_buf=full_output,
                input_scale=input_scale,
                weight_scale=weight_scale,
                output_scale=None,
                fast_accum=False,
            )
        else:
            full_output = torch.matmul(input, weight.t())
        if bias is not None:
            full_output += bias
        gemm_end_events[i].record()
        torch.distributed.reduce_scatter_tensor(output, full_output, group=TP_GROUP)
        end_events[i].record()

    gemm_times = []
    comm_times = []
    for i in range(total_iters):
        gemm_end_events[i].synchronize()
        end_events[i].synchronize()
        if i >= warmup_iters:
            gemm_times.append(start_events[i].elapsed_time(gemm_end_events[i]) / 1000)
            comm_times.append(gemm_end_events[i].elapsed_time(end_events[i]) / 1000)
    # print(gemm_times)
    # print(comm_times)
    gemm_time = sum(gemm_times) / iters * 1000
    comm_time = sum(comm_times) / iters * 1000
    return PerfResult(
        name=f"torch #{TP_GROUP.rank()}",
        output=output,
        gemm_time_ms=gemm_time,
        comm_time_ms=comm_time,
    )


@torch.no_grad()
def perf_flux(
    input: torch.Tensor,
    weight: torch.Tensor,
    bias: torch.Tensor,
    transpose_weight: bool,
    fuse_reduction: bool,
    warmup: int,
    iters: int,
    runs_per_node: bool = False,
    input_scale: Union[None, torch.Tensor] = None,
    weight_scale: Union[None, torch.Tensor] = None,
):
    is_fp8 = flux.util.is_fp8_dtype(input.dtype)
    M = input.size(0)
    # todo: transpose here to avoid TN kernel, which has the worst performence
    if transpose_weight:
        with flux.util.with_torch_deterministic(False):
            w = weight.t().contiguous()
        N = w.size(1)
    else:
        w = weight
        N = w.size(0)

    output_dtype = input.dtype if not is_fp8 else torch.bfloat16
    gemm_only_op = flux.GemmOnly(
        w.dtype,
        output_dtype,
        transpose_weight=transpose_weight,
        use_fp8_gemm=is_fp8,
    )
    cls = flux.GemmRS if not runs_per_node else flux.gemm_rs_sm80.GemmRS_multinode
    gemm_rs_op = cls(
        TP_GROUP,
        NNODES,
        M,
        N,
        input.dtype,
        output_dtype,
        transpose_weight=transpose_weight,
        fuse_reduction=fuse_reduction,
    )

    warmup_iters = warmup
    total_iters = warmup_iters + iters
    start_events = [torch.cuda.Event(enable_timing=True) for _ in range(total_iters)]
    end_events = [torch.cuda.Event(enable_timing=True) for _ in range(total_iters)]
    with flux.util.with_torch_deterministic(False):
        gemm_only_output_buf = torch.empty(
            [M, N], dtype=output_dtype, device=input.device, requires_grad=False
        )

    torch.distributed.barrier()
    for i in range(total_iters):
        start_events[i].record()
        _ = gemm_only_op.forward(
            input,
            w,
            bias=bias,
            output_buf=gemm_only_output_buf,
            input_scale=input_scale,
            weight_scale=weight_scale,
            output_scale=None,
            fast_accum=False,
        )
        end_events[i].record()
    torch.cuda.current_stream().synchronize()

    gemm_times = []
    for i in range(total_iters):
        end_events[i].synchronize()
        if i >= warmup_iters:
            gemm_times.append(start_events[i].elapsed_time(end_events[i]) / 1000)
    gemm_time = sum(gemm_times)

    time.sleep(1)

    torch.distributed.barrier()
    for i in range(total_iters):
        start_events[i].record()
        output = gemm_rs_op.forward(
            input,
            w,
            bias=bias,
            input_scale=input_scale,
            weight_scale=weight_scale,
            output_scale=None,
            fast_accum=False,
        )
        end_events[i].record()
    torch.cuda.current_stream().synchronize()

    gemm_rs_times = []
    for i in range(total_iters):
        end_events[i].synchronize()
        if i >= warmup_iters:
            gemm_rs_times.append(start_events[i].elapsed_time(end_events[i]) / 1000)
    gemm_rs_time = sum(gemm_rs_times)

    gemm_time_ms = gemm_time / iters * 1000
    comm_time_ms = (gemm_rs_time - gemm_time) / iters * 1000

    return PerfResult(
        name=f"flux  #{TP_GROUP.rank()}",
        output=output,
        gemm_time_ms=gemm_time_ms,
        comm_time_ms=comm_time_ms,
    )


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("M", type=int)
    parser.add_argument("N", type=int)
    parser.add_argument("K", type=int)
    parser.add_argument("--warmup", default=5, type=int, help="warmup iterations")
    parser.add_argument("--iters", default=100, type=int, help="perf iterations")
    parser.add_argument("--dtype", default="bfloat16", type=str, help="data type")
    parser.add_argument(
        "--profile", default=False, action="store_true", help="dump torch.profiler.profile"
    )
    parser.add_argument(
        "--transpose_weight", default=False, action="store_true", help="whether to transpose weight"
    )
    parser.add_argument(
        "--fuse_reduction", default=False, action="store_true", help="fuse reduction to gemm"
    )
    parser.add_argument("--has_bias", default=False, action="store_true", help="whether have bias")
    parser.add_argument(
        "--runs_per_node", default=False, action="store_true", help="multi-stage gemm"
    )
    return parser.parse_args()


DTYPE_MAP = {
    "bfloat16": torch.bfloat16,
    "float16": torch.float16,
}


def _rand(shape, dtype):
    if flux.util.is_fp8_dtype(dtype):
        tensor: torch.Tensor = (
            (-2 * torch.rand(shape, dtype=torch.float16, device="cuda"))
            / 100
            * (TP_GROUP.rank() + 1)
        )
        with flux.util.with_torch_deterministic(False):
            return tensor.to(dtype)
    return (-2 * torch.rand(shape, dtype=dtype).cuda() + 1) / 100 * (TP_GROUP.rank() + 1)


if __name__ == "__main__":
    torch.cuda.set_device(LOCAL_RANK)
    args = parse_args()

    print("before flux shm initialization")
    if NNODES > 1 and args.runs_per_node:
        flux.init_flux_shm(get_intra_node_pg_group(TP_GROUP, NNODES))
    else:
        flux.init_flux_shm(TP_GROUP)
    torch.cuda.synchronize()
    print("after flux shm initialization")

    dtype = DTYPE_MAP[args.dtype]
    is_fp8 = False # tobe supported in the future
    if args.transpose_weight and is_fp8:
        raise ValueError("FP8 GEMM does not support RRR layout")

    assert args.M % TP_GROUP.size() == 0
    assert args.K % TP_GROUP.size() == 0
    local_K = args.K // TP_GROUP.size()

    # input: [M, K], weight: [N, K]
    input = _rand((args.M, local_K), dtype=dtype)
    weight = _rand((args.N, local_K), dtype=dtype)

    input_scale, weight_scale = None, None


    bias = None
    bias_dtype = dtype if not is_fp8 else torch.bfloat16  # always BF16 for FP8 matmul
    if args.has_bias:
        bias = torch.rand((args.M, args.N), dtype=dtype).cuda() / 10 * (TP_GROUP.rank() + 1)

    ctx = (
        torch.profiler.profile(
            activities=[torch.profiler.ProfilerActivity.CPU, torch.profiler.ProfilerActivity.CUDA],
            record_shapes=True,
            with_stack=True,
        )
        if args.profile
        else nullcontext()
    )
    with ctx:
        perf_res_flux = perf_flux(
            input,
            weight,
            bias,
            args.transpose_weight,
            args.fuse_reduction,
            args.warmup,
            args.iters,
            args.runs_per_node,
            input_scale,
            weight_scale,
        )
        perf_res_torch = perf_torch(
            input,
            weight,
            bias,
            args.warmup,
            args.iters,
            args.transpose_weight,
            input_scale,
            weight_scale,
        )

    if args.profile:
        run_id = os.environ["TORCHELASTIC_RUN_ID"]
        prof_dir = f"prof/{run_id}"
        os.makedirs(prof_dir, exist_ok=True)
        ctx.export_chrome_trace(f"{prof_dir}/trace_rank{TP_GROUP.rank()}.json.gz")

    if TP_GROUP.rank() == 0:
        print(
            f"SOL time for GEMM(M={args.M},N={args.N},K={args.K},TP={TP_GROUP.size()}):"
            f" {flux.estimate_gemm_sol_time_ms(args.M, args.N, local_K, dtype):.3f}ms"
        )
    for i in range(TP_GROUP.size()):
        if i == TP_GROUP.rank():
            print(perf_res_torch)
        torch.distributed.barrier()
    for i in range(TP_GROUP.size()):
        if i == TP_GROUP.rank():
            print(perf_res_flux)
        torch.distributed.barrier()
    torch.distributed.barrier()

    flux_output = perf_res_flux.output
    torch_output = perf_res_torch.output
    atol = 1e-2 if dtype == torch.float16 else 2e-2
    rtol = 1e-2 if dtype == torch.float16 else 2e-2
    flux.torch_allclose(flux_output, torch_output, atol=atol, rtol=rtol)
